/**
 * \file
 * \brief KIT Hardware abstraction layer for SWI over GPIO drivers.
 *
 * \copyright (c) 2015-2020 Microchip Technology Inc. and its subsidiaries.
 *
 * \page License
 *
 * Subject to your compliance with these terms, you may use Microchip software
 * and any derivatives exclusively with Microchip products. It is your
 * responsibility to comply with third party license terms applicable to your
 * use of third party software (including open source software) that may
 * accompany Microchip software.
 *
 * THIS SOFTWARE IS SUPPLIED BY MICROCHIP "AS IS". NO WARRANTIES, WHETHER
 * EXPRESS, IMPLIED OR STATUTORY, APPLY TO THIS SOFTWARE, INCLUDING ANY IMPLIED
 * WARRANTIES OF NON-INFRINGEMENT, MERCHANTABILITY, AND FITNESS FOR A
 * PARTICULAR PURPOSE. IN NO EVENT WILL MICROCHIP BE LIABLE FOR ANY INDIRECT,
 * SPECIAL, PUNITIVE, INCIDENTAL OR CONSEQUENTIAL LOSS, DAMAGE, COST OR EXPENSE
 * OF ANY KIND WHATSOEVER RELATED TO THE SOFTWARE, HOWEVER CAUSED, EVEN IF
 * MICROCHIP HAS BEEN ADVISED OF THE POSSIBILITY OR THE DAMAGES ARE
 * FORESEEABLE. TO THE FULLEST EXTENT ALLOWED BY LAW, MICROCHIP'S TOTAL
 * LIABILITY ON ALL CLAIMS IN ANY WAY RELATED TO THIS SOFTWARE WILL NOT EXCEED
 * THE AMOUNT OF FEES, IF ANY, THAT YOU HAVE PAID DIRECTLY TO MICROCHIP FOR
 * THIS SOFTWARE.
 */

#ifndef HAL_GPIO_HARMONY
#define HAL_GPIO_HARMONY

#define __DELAY_BACKWARD_COMPATIBLE__
#include "mcc_generated_files/system/system.h"
#include "app.h"
#include <util/delay.h>
#include <string.h>

/**
 * \name Macros for Bit-Banged 1WIRE Timing
 *
 * Times to drive bits at 230.4 kbps.
   @{ */

#define tPUP                0
#define tDSCHG              150 //min spec = 150us
#define tRESET              96  //min spec = 480us (STD Speed)
#define tRRT                1   //min spec = 1ms
#define tDRR                1   //min spec = 1us; max spec = 2us
#define tMSDR               2   //min spec = 2us; max spec = 6us
#define tHTSS               150 //min spec = 150us
#define tDACK               1   //min spec = 2us; max spec = 6
/* D21 delay accuracy 1us more or equal to 2us */

#define tDACK_DLY           _delay_us(tDACK)
#define tRRT_DLY            _delay_ms(tRRT)
#define tDRR_DLY            _delay_us(tDRR)
#define tMSDR_DLY           _delay_us(tMSDR)
#define tDSCHG_DLY          _delay_us(tDSCHG)
#define tRESET_DLY          _delay_us(tRESET)
#define tHTSS_DLY           _delay_us(tHTSS)

#define tLOW0_MIN           6
#define tLOW0_MAX           16
#define tLOW1_MIN           1
#define tLOW1_MAX           2

#define tRCV_MIN            4
#define tRCV_MAX            6

#define tBIT_MIN            10
#define tBIT_MAX            75
#define tWAKEUP             1 //min spec = 1ms

#define tLOW0_TYPICAL       (tLOW0_MIN + ((tLOW0_MAX-tLOW0_MIN) / 2))   //creates typical data timing (AVG of Min and Max)
#define tLOW1_TYPICAL       (tLOW1_MIN + ((tLOW1_MAX-tLOW1_MIN) / 2))   //creates typical data timing (AVG of Min and Max)
#define tBIT_TYPICAL        (tBIT_MIN + ((tBIT_MAX-tBIT_MIN) / 2 ))     //creates typical data timing (AVG of Min and Max)

#define tLOW0_HDLY          _delay_us(11)   //min spec = 6us; max spec = 16us
#define tRD_HDLY            _delay_us(1)  //min spec = 1us; max spec = 2us
#define tLOW1_HDLY          _delay_us(1)  //min spec = 1us; max spec = 2us
#define tRCV0_HDLY          _delay_us(11)
#define tRCV1_HDLY          _delay_us(14)

#define tRD_DLY             _delay_us(1)  //min spec = 1us; max spec = 2us
#define tHIGH_SPEED_DLY     _delay_us(1)
#define tSWIN_DLY           _delay_us(1)  //delay to put master strobe within sample window

#define tLOW0_DLY           _delay_us(tLOW0_MIN)
#define tLOW1_DLY           _delay_us(tLOW1_MIN)

#define tBIT_DLY            _delay_us(tBIT_MIN)
#define tRCV0_DLY           _delay_us(tBIT_TYPICAL - tLOW0_TYPICAL)
#define tRCV1_DLY           _delay_us(tBIT_TYPICAL - tLOW1_TYPICAL)

#define send_logic0_1wire(...)      send_logic_bit(KIT_GPIO_LOGIC_BIT0)
#define send_logic1_1wire(...)      send_logic_bit(KIT_GPIO_LOGIC_BIT1)
#define send_ACK_1wire(...)         send_logic0_1wire(__VA_ARGS__)
#define send_NACK_1wire(...)        send_logic1_1wire(__VA_ARGS__)

enum kit_protocol_status hal_gpio_wake(uint32_t device_addr);
enum kit_protocol_status hal_gpio_send(uint32_t device_addr, uint8_t *txdata, uint16_t* txlength);
enum kit_protocol_status hal_gpio_receive(uint32_t device_addr, uint8_t *rxdata, uint16_t *rxlength);

#define KIT_1WIRE_RESET_WORD_ADDR              0x00
#define KIT_1WIRE_SLEEP_WORD_ADDR              0x01
#define KIT_1WIRE_SLEEP_WORD_ADDR_ALTERNATE    0x02
#define KIT_1WIRE_COMMAND_WORD_ADDR            0x03

#define KIT_1WIRE_RESPONSE_LENGTH_SIZE         0x01
#define KIT_1WIRE_BIT_MASK                     0x80
#define KIT_EMERALD_WAKE_DELAY 2
#define KIT_EMERALD_RX_RETRY 3

#define KIT_GPIO_WRITE                         0
#define KIT_GPIO_READ                          1
#define KIT_GPIO_INPUT_DIR                     0
#define KIT_GPIO_OUTPUT_DIR                    1
#define KIT_GPIO_LOGIC_BIT0                    0
#define KIT_GPIO_LOGIC_BIT1                    1
#define KIT_GPIO_ACK                           KIT_GPIO_LOGIC_BIT0
#define KIT_GPIO_CLEAR                         0
#define KIT_GPIO_SET                           1
#define KIT_MIN_RESPONSE_LENGTH                4

#define CMD_MAX_BUFFER_LEN                     40
/**
 * \name Macros for Bit-Banged SWI Timing
 *
 * Times to drive bits at 230.4 kbps.
   @{ */

//! delay macro for width of one pulse (start pulse or zero pulse)
//! should be 4.34 us, is 4.05 us

#define BIT_DELAY_1L        _delay_us(4)
//! should be 4.34 us, is 4.05us
#define BIT_DELAY_1H        _delay_us(4)

//! time to keep pin high for five pulses plus stop bit (used to bit-bang CryptoAuth 'zero' bit)
//! should be 26.04 us, is 26.92 us
#define BIT_DELAY_5        _delay_us(26)    // considering pin set delay

//! time to keep pin high for seven bits plus stop bit (used to bit-bang CryptoAuth 'one' bit)
//! should be 34.72 us, is 35.13 us
#define BIT_DELAY_7        _delay_us(34)    // considering pin set delay

//! turn around time when switching from receive to transmit
//! should be 93 us (Setting little less value as there would be other process before these steps)
#define RX_TX_DELAY        _delay_us(65)


/** SWI WORD Address */
#define KIT_SWI_WAKE_WORD_ADDR     ((uint8_t)0x00)
#define KIT_SWI_CMD_WORD_ADDR      ((uint8_t)0x77)
#define KIT_SWI_TX_WORD_ADDR       ((uint8_t)0x88)
#define KIT_SWI_IDLE_WORD_ADDR     ((uint8_t)0xBB)
#define KIT_SWI_SLEEP_WORD_ADDR    ((uint8_t)0xCC)
#define KIT_SWI_BIT_MASK           0x01

typedef enum
{
    KIT_PROTOCOL_1WIRE,
    KIT_PROTOCOL_SWI,
    NO_OF_PROTOCOL
} protocol_type;

typedef enum
{
    LOGIC0_1,
    LOGIC0_2,
    LOGIC0_3,
    LOGIC0_4,
    LOGIC1_1,
    LOGIC1_2,
    NO_OF_DELAYS
} delay_type;

enum kit_protocol_status
{
    KIT_STATUS_SUCCESS               = 0x00,
    KIT_STATUS_FAILURE               = 0x01,
    KIT_STATUS_COMMAND_NOT_VALID     = 0x03,
    KIT_STATUS_COMMAND_NOT_SUPPORTED = 0x04,
    KIT_STATUS_NO_DEVICE             = 0xC5,
    KIT_STATUS_INVALID_PARAM         = 0xE2,
    KIT_STATUS_INVALID_ID            = 0xE3,
    KIT_STATUS_INVALID_SIZE          = 0xE4,
    KIT_STATUS_RX_FAIL               = 0xE6,
    KIT_STATUS_RX_NO_RESPONSE        = 0xE7,
    KIT_STATUS_TX_TIMEOUT            = 0xEA,
    KIT_STATUS_RX_TIMEOUT            = 0xEB,
    KIT_STATUS_SMALL_BUFFER          = 0xED,
    KIT_STATUS_COMM_FAIL             = 0xF0,
    KIT_STATUS_EXECUTION_ERROR       = 0xF4,
    KIT_STATUS_TX_FAIL               = 0xF7,
};


#endif
