/* ************************************************************************** */
/** Descriptive File Name

  @Company
    Company Name

  @File Name
    filename.c

  @Summary
    Brief description of the file.

  @Description
    Describe the purpose of this file.
 */
/* ************************************************************************** */


#include "app.h"
#include "../../../project_config.h"
#include "hal_delay.h"
#include "hal_gpio_harmony.h"

/*
    Main application
*/


RESPONSE send_challenge_to_accessory(uint8_t *challenge);
RESPONSE verify_accessory_response(uint8_t chal_num);
void atCRC(size_t length, const uint8_t *data, uint8_t *crc_le);
void print_hex(uint8_t *data, size_t size);
void printHexArray(const unsigned char* array, size_t size, size_t bytesPerLine);


APP_STATE current_state;
RESPONSE status;

uint8_t packet[PACKET_SIZE+1];
uint8_t response[RESPONSE_SIZE];
uint8_t ctr[7];

extern uint8_t challenges[5][32];
extern uint8_t responses[5][32];
uint8_t chal_num;

int cr_app(void)
{
    uint8_t *challenge;
    chal_num = 0;
    while(1)
    {
        
        switch(current_state)
        {
            case CR_INIT_APP:
            {
                current_state = CR_SEND_CHALLENGE;
                break;
            }
            case CR_SEND_CHALLENGE:
            {
                challenge = challenges[chal_num];
                printf("****** Sending Challenge to Accessory Device ******\r\n\n");
                printf("Challenge being sent to accessory:\r\n");
                printHexArray(challenge, 32, 8);
                status = send_challenge_to_accessory(challenge);
                if(status == ATCA_SUCCESS)
                {
                    current_state = CR_VERIFY_RESPONSE;
                }
                else
                {
                    current_state = CR_DONE;
                }
                break;
            }
            case CR_VERIFY_RESPONSE:
            {
                printf("****** Verifying Challenge Response ******\r\n\n");
                printf("Expected accessory response:\r\n");
                printHexArray(responses[chal_num], 32, 8);
                status = verify_accessory_response(chal_num);
                current_state = CR_DONE;
                break;
            }
            case CR_DONE:
            {
                if (status != ATCA_SUCCESS)
                {
                    printf("Challenge Failed by Accessory Device\r\n");
                }
                else
                {
                    printf("Accessory Authenticated Successfully!\r\n");
                }
                while(1)
                {

                }
            }
        }
        
    }    
}

RESPONSE send_challenge_to_accessory(uint8_t *challenge)
{
    //Create the packet
    packet[0] = 0x03;
    packet[1] = 0x27;
    packet[2] = MAC_OPCODE;
    packet[3] = MAC_MODE;
    packet[4] = SYMM_SLOT;
    packet[5] = 0;
    
    memcpy(&packet[6], challenge, 32);
    //Compute CRC and add to end of packet
    uint8_t crc[2];
    atCRC(37, &packet[1], crc);
    memcpy(&packet[38], crc, 2);

#if SELECTED_SHA104_SWI
    uint16_t txlength, rxlength;
    txlength = sizeof(packet);
    rxlength = RESPONSE_SIZE;
    hal_gpio_wake(SWI_ADDR);
    hal_gpio_send(SWI_ADDR, packet, &txlength);
    atca_delay_ms(20);
    hal_gpio_receive(SWI_ADDR, response, &rxlength);
#else
    //Transmit packet to SHA104 over I2C and get response
    status = SERCOM0_I2C_Write(I2C_ADDR, packet, PACKET_SIZE + 1) ? ATCA_SUCCESS : ATCA_FUNC_FAIL;
    atca_delay_ms(16);
    do {
        atca_delay_ms(2);
        status = SERCOM0_I2C_Read(I2C_ADDR, response, RESPONSE_SIZE) ? ATCA_SUCCESS : ATCA_FUNC_FAIL;
    }while(SERCOM0_I2C_ErrorGet() != SERCOM_I2C_ERROR_NONE);

#endif
    return status;
}

RESPONSE verify_accessory_response(uint8_t chal_num)
{
    uint8_t response_data[32];
    uint8_t response_crc[2];
    memcpy(response_data, &response[1], 32);
    atCRC(33, response, response_crc);
    
    printf("Accessory response:\r\n");
    printHexArray(response_data, 32, 8);
    
    if (response_crc[0] != response[33] || response_crc[1] != response[34])
        return ATCA_CHECKMAC_VERIFY_FAILED;
    for (uint8_t i = 0; i < 32; i++)
    {
        if (response_data[i] != responses[chal_num][i])
            return ATCA_CHECKMAC_VERIFY_FAILED;
    }
    return ATCA_SUCCESS;
}

void atCRC(size_t length, const uint8_t *data, uint8_t *crc_le)
{
    size_t counter;
    uint16_t crc_register = 0;
    uint16_t polynom = 0x8005;
    uint8_t shift_register;
    uint8_t data_bit, crc_bit;

    for (counter = 0; counter < length; counter++)
    {
        for (shift_register = 0x01; shift_register > 0x00; shift_register <<= 1)
        {
            data_bit = (data[counter] & shift_register) ? 1 : 0;
            crc_bit = crc_register >> 15;
            crc_register <<= 1;
            if (data_bit != crc_bit)
            {
                crc_register ^= polynom;
            }
        }
    }
    crc_le[0] = (uint8_t)(crc_register & 0x00FF);
    crc_le[1] = (uint8_t)(crc_register >> 8);
}

void printHexArray(const unsigned char* array, size_t size, size_t bytesPerLine) {
    for (size_t i = 0; i < size; i++) {
        printf("%02X ", array[i]);
        if ((i + 1) % bytesPerLine == 0) {
            printf("\r\n");
        }
    }
    printf("\r\n");
}

/* *****************************************************************************
 End of File
 */
